"""Mysql module tests."""

from unittest import mock

import pytest
from pymysql.cursors import DictCursor

from spicerack import mysql
from spicerack.constants import PUPPET_CA_PATH


class TestMysql:
    """Mysql class tests."""

    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect_default(self, mocked_pymsql_connection):
        """It should call pymysql with the correct parameters."""
        my = mysql.Mysql(dry_run=False)
        with my.connect() as conn:
            assert conn == mocked_pymsql_connection.return_value
            call_args = mocked_pymsql_connection.call_args.kwargs
            # Ensure close is not called before context manager exits
            conn.close.assert_not_called()  # pylint: disable=maybe-no-member

        conn.close.assert_called_once_with()  # pylint: disable=maybe-no-member

        # Ensure the default args were passed
        if call_args.get("host") == "clouddb1001":
            assert call_args["read_default_group"] == "clientlabsdb"
        else:
            assert call_args["read_default_group"] == "client"

        assert call_args["charset"] == "utf8mb4"
        assert call_args["cursorclass"] == DictCursor
        assert call_args["read_default_file"].endswith("/.my.cnf")
        assert call_args["ssl"] == {"ca": PUPPET_CA_PATH}

    @pytest.mark.parametrize(
        "kwargs",
        (
            {"host": "db9999"},
            {"charset": "ascii"},
            {"charset": ""},
            {"read_default_file": "/my.cnf"},
            {"read_default_file": None, "read_default_group": None},
            {"read_default_group": "client_test"},
            {"host": "clouddb1001"},
            {"ssl": {}},
            {"ssl": {1: 3}},
        ),
    )
    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect(self, mocked_pymsql_connection, kwargs):
        """It should call pymysql with the correct parameters."""
        my = mysql.Mysql(dry_run=False)
        with my.connect(**kwargs) as conn:
            assert conn == mocked_pymsql_connection.return_value
            call_args = mocked_pymsql_connection.call_args.kwargs
            # Ensure close is not called before context manager exits
            conn.close.assert_not_called()  # pylint: disable=maybe-no-member

        conn.close.assert_called_once_with()  # pylint: disable=maybe-no-member
        # Ensure the args we passed were passed along
        for key, value in kwargs.items():
            assert call_args[key] == value

    @pytest.mark.parametrize(
        "dry_run, read_only, transaction_ro",
        (
            (False, False, False),
            (False, True, True),
            (True, False, True),
            (True, True, True),
        ),
    )
    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect_read_only(self, mocked_pymsql_connection, dry_run, read_only, transaction_ro):
        """It should start a read-only transaction if either dry-run or read-only are set."""
        my = mysql.Mysql(dry_run=dry_run)
        with my.connect(read_only=read_only) as conn:
            assert conn == mocked_pymsql_connection.return_value
            execute = conn.cursor.return_value.__enter__.return_value.execute  # pylint: disable=no-member
            print(mocked_pymsql_connection.mock_calls)
            if transaction_ro:
                execute.assert_called_once_with("SET SESSION TRANSACTION READ ONLY")
            else:
                execute.assert_not_called()



"""MySQL module (native)."""

from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

from pymysql.connections import Connection
from pymysql.cursors import DictCursor

from spicerack.constants import PUPPET_CA_PATH


class Mysql:
    """Class to manage MySQL servers.

    Caution:
        This class only has DRY-RUN support for DML sql operations.

    """

    def __init__(self, *, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            dry_run: whether this is a DRY-RUN.

        """
        self._dry_run = dry_run

    @contextmanager
    def connect(self, *, read_only: bool = False, **kwargs: Any) -> Generator:
        """Context-manager for a mysql connection to a remote host.

        Caution:
            Read-only support is limited to DML sql operations.

        Important:
            * By default autocommit is off and the commit of changes is the caller's responsibility.
            * The caller should also take care of rolling back transactions on error as appropriate.

        Arguments:
            read_only: True if this connection should use read-only transactions. **Note**: This parameter has no
                effect if DRY-RUN is set, it will be forces to True.
            **kwargs: Options passed directly to :py:class:`pymysql.connections.Connection`. If not set some settings
                will be set to default values:

                    * read_default_file: ``~/.my.cnf``. Set to :py:data:`None` to disable it.
                    * read_default_group: If not specified, it will be set to ``client`` or ``labsdbclient`` based on
                      the hostname.
                    * ssl: uses the puppet CA. Set to ``{}`` to disable.
                    * cursorclass: :py:class:`pymysql.cursors.DictCursor`

        Yields:
            :py:class:`pymysql.connections.Connection`: a context-managed mysql connection.

        Raises:
            pymysql.err.MySQLError: if unable to create the connection or set the connection to read only.

        """
        default_group = "client"
        if kwargs.get("host", "").startswith("clouddb"):
            default_group = "clientlabsdb"

        params: dict[str, Any] = {
            "charset": "utf8mb4",
            "cursorclass": DictCursor,
            "read_default_file": str(Path("~/.my.cnf").expanduser()),
            "read_default_group": default_group,
            "ssl": {"ca": PUPPET_CA_PATH},
        }
        params.update(kwargs)

        conn = Connection(**params)

        if read_only or self._dry_run:
            # FIXME: read-only support is limited to DML sql statements.
            # https://phabricator.wikimedia.org/T254756 is needed to do this better.
            with conn.cursor() as cursor:
                _ = cursor.execute("SET SESSION TRANSACTION READ ONLY")

        try:
            yield conn
        # Not catching exceptions and rolling back, as that restricts the client code
        # in how it does error handling.
        finally:
            conn.close()
